
/*
 * LIB/HISTORY.C
 *
 * (c)Copyright 1997, Matthew Dillon, All Rights Reserved.  Refer to
 *    the COPYRIGHT file in the base directory of this distribution 
 *    for specific rights granted.
 */

#include "defs.h"

Prototype void HistoryOpen(const char *fileName, int fastMode);
Prototype int  HistoryClose(void);
Prototype int HistoryLookup(const char *msgid, FILE **pfi);
Prototype int HistoryLookupByHash(hash_t hv, History *h);
Prototype int HistoryAdd(const char *msgid, History *h);
Prototype hash_t hhash(const char *msgid);
Prototype char *MsgId(char *s);

#define P1	7834891
#define P2	6017489

unsigned long	*HAry;
int		HFd = -1;
int		LoggedDHistCorrupt;
int		HFlags;

void
HistoryOpen(const char *fileName, int hflags)
{
    int fd;
    struct stat st;

    HFlags = hflags;

    if (fileName)
	fd = xopen(O_RDWR|O_CREAT, 0644, "%s", fileName);
    else
	fd = xopen(O_RDWR|O_CREAT, 0644, "%s/dhistory", NewsHome);

    if (fd < 0) {
	syslog(LOG_ERR, "open %s/dhistory failed", NewsHome);
	exit(1);
    }
    xflock(fd, XLOCK_EX);
    if (fstat(fd, &st) < 0) {
	syslog(LOG_ERR, "fstat %s/dhistory failed", NewsHome);
	exit(1);
    }
    if (st.st_size < HSIZE * sizeof(long)) {
	char *z = calloc(4096, 1);

	lseek(fd, st.st_size, 0);
	while (st.st_size < HSIZE * sizeof(long)) {
	    long n = HSIZE * sizeof(long) - st.st_size;
	    if (n > 4096)
		n = 4096;
	    if (write(fd, z, n) != n) {
		syslog(LOG_ERR, "dhistory hash table creation failed");
		exit(1);
	    }
	    st.st_size += n;
	}
	free(z);
    }

    /*
     * In FAST mode we leave the history file locked in order to
     * cache the hash table array at the beginning, which in turn
     * allows us to essentially append new entries to the end of
     * the file without having to seek back and forth updating
     * the hash table.
     *
     * When we aren't in FAST mode, we memory-map the hash table
     * portion of the history file.
     */

    if ((HFlags & HGF_FAST) == 0)
	xflock(fd, XLOCK_UN);

    if (HFlags & HGF_FAST) {
	HAry = calloc(HSIZE, sizeof(long));
	if (HAry == NULL) {
	    perror("calloc");
	    exit(1);
	}
	lseek(fd, 0L, 0);
	if (read(fd, HAry, HSIZE * sizeof(long)) != HSIZE * sizeof(long)) {
	    perror("read");
	    exit(1);
	}
    } else {
	HAry = (void *)mmap((caddr_t)0, HSIZE * sizeof(long), PROT_READ, MAP_SHARED, fd, 0);
    }

    if (HAry == NULL || HAry == (unsigned long *)-1) {
	if (fd >= 0)
	    close(fd);
	perror("mmap");
	exit(1);
    }
    HFd = fd;
}

/*
 * On close, we have to commit the hash table if we were in
 * FAST mode, otherwise we need only unmap the file before
 * closing it.
 */

int
HistoryClose(void)
{
    int r = RCOK;

    if (HFd >= 0) {
	if (HFlags & HGF_FAST) {
	    lseek(HFd, 0L, 0);
	    if (write(HFd, HAry, HSIZE * sizeof(long)) != HSIZE * sizeof(long))
		r = RCTRYAGAIN;
	    else
		free(HAry);
	} else {
	    if (HAry && HAry != (unsigned long *)-1)
		munmap((void *)HAry, HSIZE * sizeof(long));
	}
    }
    if (r == RCOK) {
	if (HFd >= 0)
	    close(HFd);
	HFd = -1;
	HAry = NULL;
    }
    return(r);
}

int
HistoryLookup(const char *msgid, FILE **pfi)
{
    hash_t hv = hhash(msgid);
    long poff = ((hv.h1 ^ hv.h2) & HMASK) * sizeof(long);
    long off = HAry[(hv.h1 ^ hv.h2) & HMASK];
    History h = { 0 };

    while (off) {
	lseek(HFd, off, 0);
	if (read(HFd, &h, sizeof(h)) != sizeof(h)) {
	    if ((LoggedDHistCorrupt & 1) == 0 || DebugOpt) {
		LoggedDHistCorrupt |= 1;
		syslog(LOG_ERR, "dhistory file corrupted on lookup");
		sleep(1);
	    }
	    break;
	}
	if (h.hv.h1 == hv.h1 && h.hv.h2 == hv.h2)
	    break;
	poff = off;
	off = h.next;
    }
    if (off != 0) {
	if (pfi) {
	    *pfi = xfopen("r", "D.%08lx/%08lx.%08lx.%04x", 
		h.gmt - h.gmt % 10, 
		hv.h1, 
		hv.h2, 
		h.iter
	    );
	}
	return(0);
    }
    return(-1);
}

int
HistoryLookupByHash(hash_t hv, History *h)
{
    long poff = ((hv.h1 ^ hv.h2) & HMASK) * sizeof(long);
    long off = HAry[(hv.h1 ^ hv.h2) & HMASK];

    while (off) {
	lseek(HFd, off, 0);
	if (read(HFd, h, sizeof(*h)) != sizeof(*h)) {
	    if ((LoggedDHistCorrupt & 1) == 0 || DebugOpt) {
		LoggedDHistCorrupt |= 1;
		syslog(LOG_ERR, "dhistory file corrupted on lookup");
		sleep(1);
	    }
	    break;
	}
	if (h->hv.h1 == hv.h1 && h->hv.h2 == hv.h2)
	    break;
	poff = off;
	off = h->next;
    }
    if (off != 0) {
	return(0);
    }
    return(-1);
}

int
HistoryAdd(const char *msgid, History *h)
{
    long hi = (h->hv.h1 ^ h->hv.h2) & HMASK;
    long poff = hi * sizeof(long);
    long off;
    int r = RCOK;

    if ((HFlags & HGF_FAST) == 0)
	xflock(HFd, XLOCK_EX);

    off = HAry[hi];

    if ((HFlags & HGF_NOSEARCH) == 0) {
	while (off) {
	    History ht;

	    lseek(HFd, off, 0);
	    if (read(HFd, &ht, sizeof(ht)) != sizeof(ht)) {
		if ((LoggedDHistCorrupt & 2) == 0 || DebugOpt) {
		    LoggedDHistCorrupt |= 2;
		    syslog(LOG_ERR, "dhistory file corrupted @ %d", off);
		    sleep(1);
		}
		break;
	    }
	    if (ht.hv.h1 == h->hv.h1 && ht.hv.h2 == h->hv.h2) {
		r = RCALREADY;
		break;
	    }
	    poff = off;
	    off = ht.next;
	}
    }

    if (r == RCOK) {
	off = lseek(HFd, 0L, 2);

	/*
	 * We insert the record into the chain at the beginning of the
	 * chain, even though we are appending the record.  This is for
	 * time-locality of reference and also reduces 'append-to-history'
	 * file overhead by localizing disk writes a bit better.
	 */

	h->next = HAry[hi];

	if (write(HFd, h, sizeof(History)) == sizeof(History)) {
	    if ((HFlags & HGF_FAST) == 0) {
		lseek(HFd, hi * sizeof(long), 0);
		if (write(HFd, &off, sizeof(long)) != sizeof(long)) {
		    ftruncate(HFd, off);
		    r = RCTRYAGAIN;
		}
	    } else {
		HAry[hi] = off;
	    }
	} else {
	    r = RCTRYAGAIN;
	}
    }

    if ((HFlags & HGF_FAST) == 0)
	xflock(HFd, XLOCK_UN);

    return(r);
}

hash_t
hhash(const char *msgid)
{
    int h1 = 0x0034629D;
    int h2 = 0x007395DD;
    int hv = 0x1A3F5C4F;
    hash_t t;

    while (*msgid) {
	h1 = h1 * *(const unsigned char *)msgid % P1;
	h2 = h2 * *(const unsigned char *)msgid % P2;
	hv = (hv << 5) ^ *(const unsigned char *)msgid ^ (hv >> 23);
	++msgid;
    }
    t.h1 = (h1 ^ (hv << 14)) & 0x7FFFFFFF;	/* bit 31 reserved for future use */
    t.h2 = (h2 ^ (hv << 20)) & 0x7FFFFFFF;	/* bit 31 reserved for future use */
    return(t);
}

char *
MsgId(char *s)
{
    int i;
    static char *LMsgId;

    if (LMsgId) {
	free(LMsgId);
	LMsgId = NULL;
    }

    if (s == NULL)
	return("<>");
    while (*s && (*s == ' ' || *s == '\t'))
	++s;
    if (*s != '<')
	return("<>");
    for (i = 0; s[i] && s[i] != '>'; ++i)
	;
    if (s[i] != '>')
	return("<>");
    if (i > MAXMSGIDLEN)
	return("<>");
    s[i+1] = 0;
    LMsgId = strcpy(malloc(strlen(s) + 1), s);
    return(LMsgId);
}

